# ============================================================================
# ・USBターゲットのデバイスオープン
# ・FPGAデバイスのコンフィギュレーション（メモリデータのWRとほぼ同じ）
# ・ハードウエアレジスタのRD/WR
# ・メモリデータのRD/WR
# ・デバイスのクローズ
# ============================================================================

# ============================================================================
# 変更履歴
# 2021/9/28     初版
# 2021/9/29     Usb_Confing() にて コンフィグデータに4MBのゼロフィルデータを
#               付け加えるようにした
# 
# ============================================================================

# ============================================================================
# import
# ============================================================================
from inspect import currentframe
import usb.core
import usb.util
import usb.backend
import const

import copy


from ctypes import c_int, c_uint8
from ctypes import c_uint16
from ctypes import c_uint32
from ctypes import c_uint64

# ============================================================================
# 定数値の設定
# ============================================================================
# USB Vendor ID
const.VID = 0x0A2F        # Prime Systems, Inc.

# USB Product ID
const.PID_KX_Card7 = 0x8001
const.PID_CX_Card5 = 0x8002
const.PID_AX_Card7 = 0x8003
const.PID_AX_Card7M = 0x8004
const.PID_SX_Card7 = 0x8005
const.PID_ZQ_Card = 0x8006

#const.USB_REQ_GET_STATUS = 0x00
#const.USB_REQ_CLEAR_FEATURE = 0x01
#const.USB_REQ_SET_FEATURE = 0x03
#const.USB_REQ_SET_ADDRESS = 0x05
#const.USB_REQ_GET_DESCRIPTOR = 0x06
#const.USB_REQ_SET_DESCRIPTOR = 0x07
#const.USB_REQ_GET_CONFIGURATION = 0x08
#const.USB_REQ_SET_CONFIGURATION = 0x09
#const.USB_REQ_GET_INTERFACE = 0x0A
#const.USB_REQ_SET_INTERFACE = 0x0B
#const.USB_REQ_SYNCH_FRAME = 0x0C
#const.USB_TYPE_STANDARD = (0x00 << 5)
#const.USB_TYPE_CLASS = (0x01 << 5)
const.USB_TYPE_VENDOR = (0x02 << 5)   # 0100_0000 = 0x40
#const.USB_TYPE_RESERVED = (0x03 << 5)
const.USB_RECIP_DEVICE = 0x00
#const.USB_RECIP_INTERFACE = 0x01
#const.USB_RECIP_ENDPOINT = 0x02
#const.USB_RECIP_OTHER = 0x03

const.USB_ENDPOINT_IN = 0x80
const.USB_ENDPOINT_OUT = 0x00

const.USB_TIME_OUT = 5000


const.SUSlv_RANGE_8BIT = 0x00  # 8bit
const.SUSlv_RANGE_16BIT = 0x40  # 16bit
const.SUSlv_RANGE_32BIT = 0x80  # 32bit
const.SUSlv_RANGE_64BIT = 0xC0  # 64bit

const.REG_ADRS_MASK = 0x003F
const.REG_RANGE_MASK = 0x00C0

# PSI社製ベンダリクエスト
const.VR_REG_TRANS_A0 = 0xa0
const.VR_REG_TRANS_B0 = 0xb0
const.VR_DATA_TRANS_A1 = 0xa1
const.VR_DATA_TRANS_B2 = 0xb2
const.VR_BOARDINFO = 0xB3  # board info

const.VR_BULK_TRANS = 0xb1
#const.VR_CONFIG_TRANS = 0xb2
#const.VR_FIRM_INFO = 0xb3

# データ転送時のコマンド
const.BULK_OUT_START = 0xC0
const.BULK_OUT_END = 0x40
const.BULK_IN_START = 0x80
const.BULK_IN_END = 0x00

# USBコンフィグ時のコマンド
const.USBCFG_START_CMD = 0xC1
const.USBCFG_END_CMD = 0x41

#const.CFGMODE_ALTERA = 0
#const.CFGMODE_XILINX = 1

# メモリコマンド
const.MEM_CMD_8BIT = 0x0000
const.MEM_CMD_16BIT = 0x0040
const.MEM_CMD_32BIT = 0x0080
const.MEM_CMD_64BIT = 0x00c0


# ============================================================================
# 範囲チェック
# ============================================================================


def isIn(_a_, _b_, _c_):
    if((((_a_) <= (_b_)) and ((_b_) <= (_c_)))):
        return True
    return False


# ============================================================================
# debug用
# ============================================================================


def dump(*args):
    names = {id(v): k for k, v in currentframe().f_back.f_locals.items()}
    print(', '.join(names.get(id(arg), '???')+' = '+repr(arg) for arg in args))


# ============================================================================
# uint8[4] <-> uint32 変換
# ============================================================================
def uint32_to_uint8(v: c_uint32):
    b: c_uint8 = [0] * 4
    b[0] = v & 0xff
    v >>= 8
    b[1] = v & 0xff
    v >>= 8
    b[2] = v & 0xff
    v >>= 8
    b[3] = v & 0xff
    return b


def uint8_to_uint32(v: c_uint8):
    ret: c_uint32 = 0

    ret |= v[3]
    ret <<= 8
    ret |= v[2]
    ret <<= 8
    ret |= v[1]
    ret <<= 8
    ret |= v[0]
    return ret


def uint16_to_uint8(v: c_uint16):
    b: c_uint8 = [0] * 2
    b[0] = v & 0xff
    v >>= 8
    b[1] = v & 0xff
    return b


def uint8_to_uint16(v: c_uint8):
    ret: c_uint16 = 0
    ret |= v[1]
    ret <<= 8
    ret |= v[0]
    return ret

# ============================================================================
# UsbOpen
# 最初に見つかった 1 台の USB デバイスをオープンします
# ============================================================================
#
# 【パラメータ】
# pid_for_open          オープンする USBデバイスの  PID
#                       0x8002  CX-Card5/7D8
#                       0x8004  AX-Card7M/20C1
#
# 【戻り値】
# usb.core.Device       オープンしたデバイスのオブジェクト
# None                  デバイスが見つかりませんでした/オープン失敗しました
#
# 【解説】
# USB VIDは 0x0A2F  (Prime Systems, Inc.) とし、
# PIDを指定してデバイスオープンします
# このあとの各種関数呼び出しには この戻り値( usb.core.Device ) を用います。
# ============================================================================


def UsbOpen(pid_for_open):

    dev = usb.core.find(idVendor=const.VID, idProduct=pid_for_open)
    if dev is None:
        print("UsbOpen() failed(1).")
        return dev  # device not found

    cfg = dev[0]
    intf = cfg[(0, 0)]

    dev.set_configuration()
    cfg = dev.get_active_configuration()
    alt = usb.util.find_descriptor(cfg, find_all=True, bInterfaceNumber=1)

    return dev

# ============================================================================
# UsbClose
# USB デバイスをクローズします
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト
#
# 【戻り値】
# None                      常にNoneを返す
#
# 【解説】
# USB デバイスをクローズします
# ============================================================================


def UsbClose(udev: usb.core.Device):
    usb.util.dispose_resources(udev)
    return None

# ============================================================================
#	各種ﾎﾞｰﾄﾞ情報の取得
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
# sel                       取得する情報
#                           0x0013 : API-Build
#                           0x0012 : API-Patch
#                           0x0011 : API-Minor
#                           0x0010 : API-Major
#                           0x00C0 : ﾌｧｰﾑｳｪｱﾊﾞｰｼﾞｮﾝ
#                           0x0080 : ﾁｯﾌﾟﾘﾋﾞｼﾞｮﾝ
#                           0x0040 : ﾌﾟﾛﾀﾞｸﾄID
#                           0x0000 : 接続ｽﾋﾟｰﾄﾞ
#
# 【戻り値】
#                          読み込んだ値
#
# 【解説】
#
#
# ============================================================================
# ============================================================================


def get_venderRequest_boardInfo(dev: usb.core.Device, sel):
    ret = dev.ctrl_transfer(
        const.USB_TYPE_VENDOR |
        const.USB_ENDPOINT_IN |
        const.USB_RECIP_DEVICE,
        const.VR_BOARDINFO,
        (sel & 0x00ff),
        0x0000,
        64
    )
    return ret

# ============================================================================
# Reg_Read
# 任意のレジスタから値を読み込みます
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
# adrs:c_uint32             対象となるレジスタ番号を指定します。
#                           0〜2,097,151 まで設定できます。
# length:c_uint32           バイト幅を指定します。1から512バイトまでの範囲です。
#
# 【戻り値】
# False                     adrs か length の範囲異常
# array.array               読み込んだ値
#
# 【解説】
# A0 コマンドを用いたレジスタ読み出しです。
# vreq_Smart-USB_Sigma.pdf 資料の bRequest=A0h 部分を参照してください。
# ============================================================================


def Reg_Read(udev: usb.core.Device, adrs: c_uint32, length: c_uint16):

    if(False == isIn(0, adrs, 0x1fffff)):
        print("Reg_Read() failed(1).")
        return False

    if(False == isIn(1, length, 512)):
        print("Reg_Read() failed(2).")
        return False

    wValue = adrs & 0xffff
    wIndex = (adrs >> 16)
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_IN | const.USB_RECIP_DEVICE,
        const.VR_REG_TRANS_A0,
        wValue,
        wIndex,
        length,
        const.USB_TIME_OUT
    )

    return ret

# ============================================================================
# Reg_Write
# 任意のレジスタに値を書き込みます。
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
# adrs:c_uint32             対象となるレジスタ番号を指定します。
#                           0〜2,097,151 まで設定できます。
# data:bytearray            書き込むデータ(little endian)
#
# 【戻り値】
# False                     範囲チェックやデータサイズでエラーがあります
# 数値                      書き込んだサイズ
#
# 【解説】
# A0 コマンドを用いたレジスタ書き込みです。
# vreq_Smart-USB_Sigma.pdf 資料の bRequest=A0h 部分を参照してください。
# ============================================================================


def Reg_Write(udev: usb.core.Device, adrs: c_uint32, data: bytearray):

    # 範囲チェック
    if(False == ((0 <= adrs) and (adrs <= 0x1fffff))):
        print("Reg_Write() failed(1).")
        return False

    length = len(data)
    if(False == ((1 <= length) and (length <= 512))):
        print("Reg_Write() failed(2).")
        return False

    wValue = adrs & 0xffff
    wIndex = (adrs >> 16)

    # print(wValue)
    # print(wIndex)
    # print(data)

    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,
        const.VR_REG_TRANS_A0,
        wValue,
        wIndex,
        data,
        const.USB_TIME_OUT
    )

    return ret


# ============================================================================
# Reg_Read_B0
# B0コマンドを用いたレジスタ読み出し
# ============================================================================
#
# 【パラメータ】
# udev                      Open でオープンしたオブジェクト
# adrs:c_uint32             対象となるレジスタ番号を指定します。
#                           0〜2,097,151 まで設定できます。
# SUSlv_range               アクセス幅を指定します。下記が入ります。
#                           const.SUSlv_RANGE_8BIT    8bit幅
#                           const.SUSlv_RANGE_16BIT  16bit幅
#                           const.SUSlv_RANGE_32BIT  32bit幅
#                           const.SUSlv_RANGE_64BIT  64bit幅#
# 【戻り値】
# False                     異常な返り値が得られた
# array.array               読み込んだ値の配列
#
# 【解説】
# 1つのレジスタを 8/16/32/64ビット幅で読み出します。
# ============================================================================
def Reg_Read_B0(udev, adrs: c_uint32, SUSlv_range):

    cmd: c_uint16 = 0x0000
    length: c_uint16 = 0
    if(SUSlv_range == const.SUSlv_RANGE_16BIT):
        cmd = const.MEM_CMD_16BIT
        length = 2
    elif(SUSlv_range == const.SUSlv_RANGE_32BIT):
        cmd = const.MEM_CMD_32BIT
        length = 4
    elif(SUSlv_range == const.SUSlv_RANGE_64BIT):
        cmd = const.MEM_CMD_64BIT
        length = 8
    else:
        # const.SUSlv_RANGE_8BIT
        cmd = const.MEM_CMD_8BIT
        length = 1

    adrs &= const.REG_ADRS_MASK
    cmd |= adrs

    #print("length = " , length)
    #print("cmd    = 0x{:X}".format(cmd))

    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_IN | const.USB_RECIP_DEVICE,
        const.VR_REG_TRANS_B0,
        cmd,
        0x0000,
        length,
        const.USB_TIME_OUT
    )

    # print(ret)

    ret_len = len(ret)
    if(ret_len == 1):
        result: c_uint8 = 0
    elif(ret_len == 2):
        result: c_uint16 = 0
    elif(ret_len == 4):
        result: c_uint32 = 0
    elif(ret_len == 8):
        result: c_uint64 = 0
    else:
        return False

    for i in range(ret_len):
        result <<= 8
        result |= ret[ret_len-1-i]

    return result


# ============================================================================
# Reg_Write_B0
# 任意のレジスタに値を書き込みます。
# ============================================================================
#
# 【パラメータ】
# udev                      Open でオープンしたオブジェクト
# adrs:c_uint32             対象となるレジスタ番号
#                           0~2,097,151 まで設定できます。
# SUSlv_range               アクセス幅を指定します。下記が入ります。
#                           const.SUSlv_RANGE_8BIT    8bit幅
#                           const.SUSlv_RANGE_16BIT  16bit幅
#                           const.SUSlv_RANGE_32BIT  32bit幅
#                           const.SUSlv_RANGE_64BIT  64bit幅
# data                      書き込むデータ()
#
# 【戻り値】
# False                     書き込み失敗
# 数値                      書き込み数（バイト単位）
#
# 【解説】
#
# ============================================================================
def Reg_Write_B0(udev, adrs: c_uint32, SUSlv_range, data):

    cmd: c_uint16 = 0x0000
    length = 0
    if(SUSlv_range == const.SUSlv_RANGE_16BIT):
        cmd = const.MEM_CMD_16BIT
        length = 2
    elif(SUSlv_range == const.SUSlv_RANGE_32BIT):
        cmd = const.MEM_CMD_32BIT
        length = 4
    elif(SUSlv_range == const.SUSlv_RANGE_64BIT):
        cmd = const.MEM_CMD_64BIT
        length = 8
    else:  # SUSlv_range == SUSlv_RANGE_8BIT or default
        cmd = const.MEM_CMD_8BIT
        length = 1

    adrs &= const.REG_ADRS_MASK
    cmd |= adrs

    cntrol_msg = list(range(length))
    for i in range(length):
        cntrol_msg[i] = data & 0xff
        data >>= 8
    
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,
        const.VR_REG_TRANS_B0,
        cmd,
        0x0000,
        cntrol_msg,
        const.USB_TIME_OUT
    )

    return ret

# ============================================================================
# Data_Write
# メモリに値を書き込みます。
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
# data:bytearray            書き込むデータ(little endian)
#
# 【戻り値】
# False                     範囲チェックやデータサイズでエラーがあります
# 数値                      書き込んだサイズ
#
# 【解説】
# vreq_Smart-USB_Sigma.pdf 資料の bRequest=A1h 部分を参照してください。
# ============================================================================


def Data_Write(udev: usb.core.Device, data: bytearray):

    MAX_transferSize_for_write_uint8 = 0x100000

    # USBバージョンによる 最大書き込みサイズの設定
    if(udev.speed == 2):    # USB1.1
        min_transferSize = 2048
    elif(udev.speed == 3):  # USB2.0
        min_transferSize = 16384
    elif(udev.speed == 4):  # USB3.0
        min_transferSize = 32768
    else:
        print("Data_Write() failed(1).")
        print("error: USB device speed = ", udev.speed)
        return False

    # 書き込むデータのサイズ
    dataSize_uint8 = len(data)

    # 最小アクセス単位を考慮した 書き込むデータサイズ
    to_transfer_size_uint8 = dataSize_uint8
    _div = dataSize_uint8 // min_transferSize
    _mod = dataSize_uint8 % min_transferSize
    if(dataSize_uint8 % min_transferSize):
        to_transfer_size_uint8 = (_div+1) * min_transferSize

    # FPGAに通知する uint32(4byte単位) での書き込むカウント
    dataSize_uint32 = dataSize_uint8 // 4
    if(dataSize_uint8 % 4):
        dataSize_uint32 += 1

    # 範囲チェック
    if(False == ((0 <= dataSize_uint32) and (dataSize_uint32 <= 0x3fffffff))):
        print("Data_Write() failed(2).")
        return False

    param: c_uint32 = (1 << 31) | (1 << 30) | dataSize_uint32
    wValue: c_uint16 = param & 0xffff
    wIndex: c_uint16 = (param >> 16)
    #print("wValue = 0x{:04X}" . format(wValue))
    #print("wIndex = 0x{:04X}" . format(wIndex))
    #print("dataSize_uint32 = ", dataSize_uint32)
    #print("dataSize_uint8 = ", dataSize_uint8)

    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_A1,  # 0xa1
        wValue,
        wIndex,
        0,
        const.USB_TIME_OUT
    )
    if(ret == None):
        print("Data_Write() failed(3).")
        return False

    transfered_byte = 0
    atleast = dataSize_uint8

    idx_begin = 0
    idx_end = 0

    # min_transferSize の倍数の転送サイズで書き込みする
    while(0 < atleast):
        # 転送サイズは 最大値にしておく
        transfer_size = MAX_transferSize_for_write_uint8

        # 転送残量より小さければ 転送サイズは 転送残量とする
        if(atleast < transfer_size):
            transfer_size = atleast

            # 転送サイズが最低転送量で割り切れない場合は まず最低転送量で割り切れる転送量にする
            if(0 < (transfer_size % min_transferSize)):
                tmp = transfer_size // min_transferSize
                transfer_size = tmp * min_transferSize

            # 転送ザイズが 最低転送量より小さければ 転送サイズは 最低転送量とする
            if(transfer_size < min_transferSize):
                transfer_size = min_transferSize

        siz = transfer_size
        idx_end = idx_begin + siz

        # atleastがまだ残っているならば dataのスライス、残っていないなら コピーを作る
        buf = data[idx_begin:idx_end]
        blank = [0] * (siz-atleast)
        buf.extend(blank)

        ret = udev.write(0x6, buf, const.USB_TIME_OUT)
        if(ret == None):
            print("Data_Write() failed(4).")
            return transfered_byte
        transfered_byte += ret
        atleast -= ret
        if(atleast < 0):
            atleast = 0
        idx_begin += ret

        # 書き込みの様子を表示する場合は 下記をコメントアウトする
        # print("write 0x%08x to 0x%08x ... %9d(0x%08x) bytes read , atlest %9d(0x%08x) bytes" % (
        #    idx_begin, idx_begin+transfer_size-1, transfer_size, transfered_byte, atleast, atleast))

    param: c_uint32 = (0 << 31) | (1 << 30) | dataSize_uint32
    #param = (0<<31) | (1<<30)
    wValue = param & 0xffff
    wIndex = (param >> 16)
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_A1,  # 0xa1
        wValue,
        wIndex,
        0,
        const.USB_TIME_OUT
    )

    return transfered_byte


# ============================================================================
# Data_Read
# メモリから値を読み込みます
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
# dataSize_uint8:c_uint32   読み込むバイト数　4の倍数以外も許容します
#
# 【戻り値】
# False                     範囲チェックやデータサイズでエラーがあります
# 数値                      書き込んだサイズ
#
# 【解説】
# vreq_Smart-USB_Sigma.pdf 資料の bRequest=A1h 部分を参照してください。
# ============================================================================


def Data_Read(udev: usb.core.Device, dataSize_uint8: c_uint32):

    # 一度に読み込める最大サイズの指定
    # 0x0100_0000 だと usb.core.USBError: [Errno 5] Input/Output Error がでた
    MAX_transferSize_for_read_uint8 = 0x00800000

    # USBバージョンによる 読み込みサイズの設定
    if(udev.speed == 2):      # USB1.1
        min_transferSize = 256
    elif(udev.speed == 3):    # USB2.0
        min_transferSize = 2048
    elif(udev.speed == 4):    # USB3.0
        min_transferSize = 4096
    else:
        print("Data_Read() failed(1).")
        print("error: USB device speed = ", udev.speed)
        return False

    # 最小アクセス単位を考慮した 読み込むデータサイズ
    to_transfer_size_uint8 = dataSize_uint8
    if(dataSize_uint8 % min_transferSize):
        _div = dataSize_uint8 // min_transferSize
        to_transfer_size_uint8 = (_div+1) * min_transferSize

    # FPGAに通知する uint32(4byte単位) での読み込みカウント
    dataSize_uint32 = dataSize_uint8 // 4
    if(dataSize_uint8 % 4):
        dataSize_uint32 += 1

    to_transfer_size_uint32 = to_transfer_size_uint8 // 4

    #print("dataSize_uint8           = ", dataSize_uint8)
    #print("to_transfer_size_uint8   = ", to_transfer_size_uint8)
    #print("to_transfer_size_uint32  = ", to_transfer_size_uint32)

    # 範囲チェック
    if(False == ((0 <= to_transfer_size_uint32) and (to_transfer_size_uint32 <= 0x3fffffff))):
        print("Data_Read() failed(2).")
        return False

    # 転送開始・メモリリード
    param: c_uint32 = (1 << 31) | (0 << 30) | to_transfer_size_uint32
    wValue = param & 0xffff
    wIndex = (param >> 16)

    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_A1,  # 0xa1
        wValue,
        wIndex,
        0,
        const.USB_TIME_OUT
    )
    if(ret == None):
        print("Data_Read() failed(3).")
        return False

    transfered_byte = 0
    atleast = dataSize_uint8
    data = []

    idx_begin = 0

    while(0 < atleast):

        # 転送サイズは 最大値にしておく
        transfer_size = MAX_transferSize_for_read_uint8

        # 転送残量より小さければ 転送サイズは 転送残量とする
        if(atleast < transfer_size):
            transfer_size = atleast

            # 転送サイズが最低転送量で割り切れない場合は まず最低転送量で割り切れる転送量にする
            if(0 < (transfer_size % min_transferSize)):
                tmp = transfer_size // min_transferSize
                transfer_size = (tmp+1) * min_transferSize

            # 転送ザイズが 最低転送量より小さければ 転送サイズは 最低転送量とする
            if(transfer_size < min_transferSize):
                transfer_size = min_transferSize

        # EP2(0x82) から読み出し
        ret = udev.read(0x82, transfer_size, const.USB_TIME_OUT)
        if(ret == None):
            print("Data_Read() failed(4).")
            return transfered_byte

        data.extend(ret)
        transfered_byte += len(ret)

        atleast -= transfer_size
        if(atleast < 0):
            atleast = 0
        # print("read 0x%08x to 0x%08x ... %9d(0x%08x) bytes read , atlest %9d(0x%08x) bytes" % (
        #    idx_begin, idx_begin+transfer_size-1, transfer_size, transfered_byte, atleast, atleast))
        idx_begin += transfer_size

    # 転送終了・メモリリード
    param: c_uint32 = (0 << 31) | (0 << 30) | to_transfer_size_uint32
    wValue = param & 0xffff
    wIndex = (param >> 16)
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_A1,  # 0xa1
        wValue,
        wIndex,
        0,
        const.USB_TIME_OUT
    )

    del data[dataSize_uint8:]
    return data, transfered_byte


# ============================================================================
# Usb_Config
# FGPAをコンフィグします
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
# configFilename            コンフィグファイル名
#
# 【戻り値】
# False                     範囲チェックやデータサイズでエラーがあります
# 数値                      書き込んだサイズ
#
# 【解説】
#
# ============================================================================

def Usb_Confing(udev: usb.core.Device, configFilename, fpga):

    with open(configFilename, "rb") as f:
        dataF = f.read()
    orig = bytearray(dataF)
    length_of_uint8 = len(orig)
    f.close()

    # USB productID を取得
    ret = get_venderRequest_boardInfo(udev, 0x0040)
    productID = ret[1] << 8 | ret[0]

    # データ列を変換 0;Xilinx  1;Altera
    if(fpga == 1):
        configData = dataconvert_to_CXcard5(orig, length_of_uint8)
    elif(fpga == 0):
        configData = dataconvert_defaultCard(orig, length_of_uint8)


    # 4MB程度 余計に流し込むため コンフィグデータに0データらを追加する
    StartupSize = 4 * 1024 * 1024
    StartupBuf: c_uint8 = [0] * StartupSize
    configData = configData + StartupBuf

    # コンフィグ開始を送信
    cmd = [const.USBCFG_START_CMD]
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_B2,  # 0xb2
        0x0000,
        0x0000,
        cmd,
        const.USB_TIME_OUT
    )
    if(ret == None):
        print("Usb_Confing() failed(1).")
        return False

    # コンフィグデータをバルク転送する
    ret = Data_Write(udev, configData)

    # コンフィグステータス確認
    reg_addr = 0
    ret = Reg_Read_B0(udev, reg_addr, const.SUSlv_RANGE_8BIT)
    if(ret != 0):
        # 書き込み失敗
        print("Usb_Confing() failed(3).ret=", ret)
        return False

    # コンフィグ終了を送信
    cmd = [const.USBCFG_END_CMD]
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_B2,  # 0xb2
        0x0000,  # config end
        0x0000,
        cmd,
        const.USB_TIME_OUT
    )
    if(ret == None):
        print("Usb_Confing() failed(2).")
        return False

    return True


# ============================================================================
# Usb_Config 用 コンフィグデータのフォーマット変更
# CXcard5 用
# ============================================================================
def dataconvert_to_CXcard5(orig, length_of_uint8):
    configData: c_uint8 = [0] * (length_of_uint8 * 8 + 1)
    j = 0
    for i in range(0, length_of_uint8-1, 2):
        configData[j + 0] = orig[i+0]
        configData[j + 1] = orig[i+1]
        configData[j + 2] = 0
        configData[j + 3] = 0
        configData[j + 4] = orig[i+0]
        configData[j + 5] = orig[i+1]
        configData[j + 6] = 0
        configData[j + 7] = 0
        configData[j + 8] = orig[i+0]
        configData[j + 9] = orig[i+1]
        configData[j+10] = 0
        configData[j+11] = 0
        configData[j+12] = orig[i+0]
        configData[j+13] = orig[i+1]
        configData[j+14] = 0
        configData[j+15] = 0
        j += 16
    return configData


# ============================================================================
# Usb_Config 用 コンフィグデータのフォーマット変更
# CXcard5 以外
# ============================================================================
def dataconvert_defaultCard(orig, length_of_uint8):

    configData: c_uint8 = [0] * (length_of_uint8 * 4 + 1)
    j = 0
    for i in range(0, length_of_uint8-1):
        configData[j + 0] = orig[i]
        configData[j + 1] = 0
        configData[j + 2] = 0
        configData[j + 3] = 0
        j += 4
    return configData


# ============================================================================
# Usb_Deconfig
# FGPAを未コンフィグ状態にします
# ============================================================================
#
# 【パラメータ】
# udev:usb.core.Device      Open でオープンしたオブジェクト()
#
# 【戻り値】
# False                     範囲チェックやデータサイズでエラーがあります
# 数値                      書き込んだサイズ
#
# 【解説】
#
# ============================================================================


def Usb_Deconfing(udev: usb.core.Device):

    cmd = [const.USBCFG_START_CMD]
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_B2,  # 0xb2
        0x0000,
        0x0000,
        cmd,
        const.USB_TIME_OUT
    )
    if(ret == None):
        print("Usb_Deconfing() failed(1).")
        return False

    # なにも書き込まないで コンフィグを終了する

    cmd = [const.USBCFG_END_CMD]
    ret = udev.ctrl_transfer(
        const.USB_TYPE_VENDOR | const.USB_ENDPOINT_OUT | const.USB_RECIP_DEVICE,  # 0x40
        const.VR_DATA_TRANS_B2,  # 0xb2
        0x0000,  # config end
        0x0000,
        cmd,
        const.USB_TIME_OUT
    )
    if(ret == None):
        print("Usb_Deconfing() failed(2).")
        return False

    return True
